#!/usr/bin/env python
# -*- coding: UTF-8 -*-
"""Import list of currencies from XML file provided by Federal Reserve Bank NY.

Download http://www.unece.org/etrades/unedocs/repository/codelists/xml/CurrencyCodeList.xml and give the file as first command line argument.
"""

__all__ = ['import_currencies', 'main', 'parse_currency_xml']

__program__   = "import_currencies"
__author__    = "Christopher Arndt"
__revision__  = "$Rev$"
__date__      = "$Date$"
__license__ = "LGPL"

import sys
import logging

from optparse import OptionParser
from xml.etree import ElementTree

import turbogears

from sqlobject import SQLObjectNotFound

from luca.util import fxrates_client
from luca.util.fxrates_client import *

def usage():
    optparser.print_help()

def parse_currency_xml(file):
    doc = ElementTree.parse(file)
    cl = doc.findall('Currency')

    l = []

    for curr in cl:
        d = {}
        d['name'] = curr.find('CurrencyName').text.strip()
        d['code'] = curr.find('CurrencyCoded').text.strip()
        l.append(d)

    l.sort(key=lambda x: x['code'])
    return l

def import_currencies(xmlfile, check_rates=True, verbose=False, dryrun=False):
    """Parse currency list from XML file and enter into database."""

    from luca.model import Currency, CurrencyRate, hub

    if verbose:
        logging.info("Reading currency list from '%s'." % xmlfile)
    try:
        currencies = parse_currency_xml(xmlfile)
    except Exception, exc:
        logging.error(exc)
        return 1

    if verbose:
        logging.info("Creating currency and currency_rates database tables.")
    if not dryrun:
        Currency.createTable(ifNotExists=True)
        CurrencyRate.createTable(ifNotExists=True)

    rates = {}
    if check_rates:
        if verbose:
            logging.info("Retrieving FX rates from FRBNY.")
        try:
            rates = get_all_latest_noon_rates()
        except (fxrates_client.ParseError, fxrates_client.NetworkError), exc:
            logging.error(exc)

    hub.begin()
    for curr in currencies:
        auto = curr['code'] in rates
        try:
            cobj = Currency.byCode(curr["code"])
        except SQLObjectNotFound:
            if verbose:
                logging.info("Importing new currency: '%s' (auto download=%r)" %
                  (curr['code'], auto))
            if not dryrun:
                Currency(name=curr['name'], code=curr['code'], autoDownload=auto)
        else:
            logging.info("Currency %s already in database (name=%s)" %
              (cobj.code, cobj.name))
            if cobj.autoDownload != auto:
                if verbose:
                    logging.info("Activating autoDownload for currency %s"
                        % (cobj.code,))
                if not dryrun:
                    cobj.autoDownload = auto
    hub.commit()
    return 0


def main(args=None):
    global optparser

    if args is None:
        args = sys.argv[1:]

    optparser = OptionParser(prog=__program__,
      version=__revision__, description=__doc__)
    optparser.add_option("-v", "--verbose",
      action="store_true", dest="verbose", default=False,
      help="Report what's going on to stdout.")
    optparser.add_option("-n", "--dry-run",
      action="store_true", dest="dryrun", default=False,
      help="Don't actually write to the database (implies -v).")
    optparser.add_option("-c", "--config", metavar='CONFIG',
      action="store", dest="config", default='dev.cfg',
      help="Read database configuration from CONFIG.")
    optparser.add_option("-r", "--check-rates",
      action="store_true", dest="check_rates", default=False,
      help="Check for availability of exchange rates from FRBNY (needs internet conection).")

    (options, args) = optparser.parse_args(args=args)

    if options.dryrun:
        options.verbose = True
    if options.verbose:
        logging.basicConfig(level=logging.INFO)
    if not args:
        usage()
        return 2

    if options.verbose:
        logging.info("Reading configuration from '%s'." % options.config)
    turbogears.update_config(configfile=options.config,
      modulename="luca.config")

    return import_currencies(args[0], options.check_rates, options.verbose,
      options.dryrun)


if __name__ == '__main__':
    sys.exit(main())
