# -*- coding: UTF-8 -*-
"""Controller for the foreign currency rates table."""

__all__ = ['currency_rates']

__author__    = "Christopher Arndt"
__revision__  = "$Rev$"
__date__      = "$Date$"
__license__   = "LGPL"

import datetime

import cherrypy
import turbogears

from turbogears import (config, controllers, error_handler, expose, flash,
  identity, redirect, url, validate)
from turbogears import validators as vldtrs

from sqlobject import sqlbuilder

from luca.model import Currency, CurrencyRate, conn, notfound, denied
from luca import model

from luca.simple_table import simple_table
from mylib import date_formats, thousand_separators, decimal_separators

from luca.widgets.currencies import *


class currency_rates(simple_table):

    # SUBCLASS-SPECIFIC FUNCTIONS AND PARAMETERS ##########################

    validators = {"id": vldtrs.Int(if_invalid=0),
                  "fxrate": DecimalValidator(if_invalid=-1),
                  "date": DateValidator(if_invalid=-1),
                  "manualOverride": vldtrs.Int(if_invalid=1),
                  "currency": vldtrs.Int(if_invalid=-1),
                  "lite":  vldtrs.Int(if_invalid=0) \
                }

    def setup(self):

        # data format for building JSON responses
        self.emptyrow_read = {"id": 0, "fxrate": 1.0, "date": "",
            "manualOverride": 1, "currency": None}
        self.emptyrow_write = {"id": 0, "fxrate": 1.0, "date": "",
            "manualOverride": 1, "currency": None, 'user': ''}
        self.empty_intermediate = self.emptyrow_read

        self.formatters = dict(manualOverride=self.bool2string)

        # data for automated tests
        self.sample_data = [{},{},{},{}]
        self.invalid_data = [("fxrate", ""), ("date", ""),
          ("manualOverride", "")]
        self.invalid_compound_new = []
        self.unique_cols = []
        self.foreign_keys = ["currency"]
        self.test_order = 9
        self.test_flags = ('cant_remove_id_1',)

        # error messages
        self.msg_unexpected = "ERROR: unexpected SQL error. Contact the system administrator"
        self.msg_notfound = "Currency not found, probably it has just been deleted"
        self.show_msg_ajax = False
        self.msg_added = "FX rate %d has been added"
        self.msg_updated = "FX rate has been updated"
        self.msg_removed = "FX rate has been removed sucessfully"
        self.msg_validation = "FX rate and date must not be empty"
        self.msg_validation_crash = "Unexpected validation error: notify system administrator!"
        self.msg_integrity = "Integrity error: "
        self.msg_cannotremove = "FX rate can not be removed"
        self.msg_unique = "Make sure the currency/date combination is unique"
        self.msg_legend_update = "FX rates: editing FX rate for '%s'"
        self.msg_legend_add = "FX rates: adding new FX rate"

        self.table = CurrencyRate
        self.streamlined_add = True

        self.lite_options = [
            {"id": 0, "description": u"AJAX menu by default"},
            {"id": 1, "description": u"'Lite' (no Javascript) menu by default"}
        ]

    def data_legend(self, d):
        return d["date"]

    def row_name(self, d_row):
        name = ""
        if d_row["id"] > 0:
            name = u"%s - %s (id %d)" % (d_row["fxrate"], d_row["date"],
              d_row["id"])
        else:
            name = u"%.2f - %s" % (d_row["fxrate"], d_row["date"])
        return name

    def auxiliary_form_data(self, form_data):
        form_data["lite_options"] = self.lite_options

    def intermediate_to_form_ext(self, intermediate, form_data):
        """Process outgoing data for form display."""

        if intermediate["currency"]:
            if hasattr(intermediate["currency"], 'id'):
                form_data["currency"] = intermediate["currency"].id
            else:
                form_data["currency"] = intermediate["currency"]
        form_data["manualOverride"] = 1

    def form_to_intermediate_ext(self, form_data, intermediate):
        """Process incoming from data."""

        try:
            intermediate["currency"] = Currency.get(form_data["currency"])
        except notfound:
            intermediate["currency"] = -1
        intermediate["user"] = identity.current.user

    def intermediate_validate(self, d):
        err = ""
        if not d["date"]:
            err += "FX rate date must not be empty\n"
        if not d["fxrate"]:
            err += "FX rate must not be empty\n"
        elif d["fxrate"] == -1:
            err += "Invalid FX rate value\n"
        if not d["date"]:
            err += "Date must not be empty\n"
        elif d["date"] == -1:
            err += "Invalid date string (format YYYY-MM-DD)\n"
        if not d["currency"]:
            err += "Missing currency ID\n"
        if d["currency"] == -1:
            err += "Currency not found\n"
        # unique currency/date combination?
        if not err:
            if d["id"] < 0:
                res = self.table.select(sqlbuilder.AND(
                    self.table.q.date == d["date"],
                    self.table.q.currencyID == d["currency"].id
                ))
            else:
                res = self.table.select(sqlbuilder.AND(
                    self.table.q.date == d["date"],
                    self.table.q.currencyID == d["currency"].id,
                    self.table.q.id != d["id"]
                ))
            if res.count() > 0:
                err += self.msg_integrity + self.msg_unique

        return err

    def rates_by_currency(self, code, tg_errors=None):
        if tg_errors:
            flash(u'Invalid currency code.')
            redirect('/currencies/')
        try:
            curr = Currency.byCode(code)
        except notfound:
            flash(u'Currency not found.')
            redirect('/currencies/')
        data = self.base_index(self.table.q.currencyID == curr.id,
          orderBy=sqlbuilder.DESC(self.table.q.date))
        data["currency"] = curr
        data["download_rates_form"] = download_rates_form
        data["download_rates_form_data"] = dict(
          action=url('/currency_rates/download_rates'),
          submit_text=_(u'Start download'),
          value = dict(
            startdate=datetime.date.today() - datetime.timedelta(7),
            code=curr.code))
        return data

    # EXPOSED METHODS #####################################################

    @expose(template="luca.templates.currency_rates")
    @validate(validators=dict(code=vldtrs.String(max=3)))
    def index(self, code, tg_errors=None):
        return self.rates_by_currency(code)

    @expose(template="luca.templates.currency_rates_lite")
    def lite(self, **garbage):
        return self.rates_by_currency(code)

    @expose("json")
    @validate(validators=validators)
    def remove(self, id, **garbage):
        return self.base_remove(id)

    @expose()
    @validate(validators=validators)
    def lite_remove(self, id, **garbage):
        return self.base_lite_remove(id)

    @expose("json")
    @validate(validators=validators)
    def fetch(self, id, currency=None, **kw):
        self.log.debug('Fetch data: %r' % kw)
        if currency:
            return self.base_fetch(id, dict(currency=currency))
        return self.base_fetch(id)

    @expose(template="luca.templates.currency_rates_lite_form")
    @validate(validators=validators)
    def lite_edit(self, id, **garbage):
        return self.base_lite_edit(id)

    @expose("json")
    @validate(validators=validators)
    def update(self, **form_data):
        self.log.debug('Update data: %r' % form_data)
        return self.base_update(form_data)

    @expose(template="luca.templates.currency_rates_lite_form")
    @validate(validators=validators)
    def lite_commit(self,  **form_data):
        return self.base_lite_commit(form_data)

    @expose(template="luca.templates.download_rates", allow_json=True)
    @validate(validators=DownloadRatesSchema())
    def download_rates(self, code=None, startdate=None, enddate=None,
      keep_manual=True, tg_errors=None):
        d = dict()
        if tg_errors:
            d['err'] = [unicode(e).encode('utf-8') for e in
              tg_errors.values()]
        else:
            if code:
                try:
                    curr = Currency.byCode(code)
                except notfound:
                    d['err'] = [u'Currency not found']
                    return d
                else:
                    currencies = [curr]
                    d['currency_code'] = curr.code
            else:
                currencies = Currency.select()
            d['codes'] = []
            d['msg'] = ['Starting download of currency exchange rates.']
            for curr in currencies:
                if not curr.autoDownload:
                    d['msg'].append(u'Not downloading rates for %s' % curr.code)
                    continue
                # FIXME: Where to set the key currency?
                if curr.code == config.get('luca.fxrates.key_currency', 'USD'):
                    continue
                try:
                    d['msg'].extend(curr.download_rates(startdate, enddate,
                      keep_manual))
                except Exception, e:
                    d['msg'].append(u'Error downloading rates for %s: %s' % \
                      (curr.code, e))
                # do updates
                d['codes'].append(curr.code)
            else:
                d['msg'].append(u'Download of currency exchange rates completed.')
        return d

    # formatter methods
    def bool2string(self, val):
        if val:
            return "Yes"
        return "No"
