from luca.model import *
import mylib
import accountlib
from sqlobject import *

# These functions do not begin/commit transactions, nor they handle SQL
# errors, neither sort the results. These should be handled by the caller.

def account_balance(org_id, account_id, date):
	'''Calculates the account balance at the BEGINNING of the given day'''

	gconn.cache.clear()

	filter = []

	filter.append(sqlbuilder.SQLConstant("journal.cancelled <= 0"))
	filter.append(sqlbuilder.SQLConstant("journal.org_id = %d" % org_id))
	filter.append(sqlbuilder.SQLConstant("journal.account_id = %d" % account_id))
	filter.append(sqlbuilder.SQLConstant("journal.tday < '%s'" % date))

	if cmos.get(1).batch_open == 0:
		filter.append(batches.q.closed > 0)

	res = journal.select(sqlbuilder.AND(*filter), \
			join=sqlbuilder.INNERJOINOn(None, batches, sqlbuilder.SQLConstant("journal.batch_id = batches.id")))

	total = res.accumulate('SUM(journal.tvalue)')

	return total and int(total) or 0

def balance_report(org_id, date1, date2, direction, noadjust, result, normal_balance):
	'''Generates an account report. Suitable for: balances, income statements and cash flow statements

		direction: 0=both, 1=credits only, 2=debits only
		noadjust: 0=whatever, 1=no
		result: 0=all accounts, 1=only patrimonial accounts 2=only income (revenue/expense)
			(The "result" name is due to a previous bad translation of the "Income" term
				from Portuguese to English)
		normal_balance: 0=whatever, 1=cr only, 2=db only

		returned list of dicts is NOT sorted in any way
	'''

	gconn.cache.clear()

	dummy, err = accountlib.check_chart(orgs.get(org_id).chart, cmos.get(1).accounts_checkdigit)
	if err:
		return None

	dbfilter = []
	dbfilter.append(sqlbuilder.SQLConstant("journal.cancelled <= 0"))
	dbfilter.append(sqlbuilder.SQLConstant("journal.org_id = %d" % org_id))

	if cmos.get(1).batch_open == 0:
		dbfilter.append(sqlbuilder.SQLConstant("batches.closed > 0"))

	if direction == 1:
		dbfilter.append(sqlbuilder.SQLConstant("journal.value > 0"))
	elif direction == 2:
		dbfilter.append(sqlbuilder.SQLConstant("journal.value < 0"))

	if noadjust:
		dbfilter.append(sqlbuilder.SQLConstant("batches.adjust <= 0"))

	crfilter = dbfilter[:]
	inbalfilter = dbfilter[:]

	inbalfilter.append(sqlbuilder.SQLConstant("journal.tday < '%s'" % date1))

	dbfilter.append(sqlbuilder.SQLConstant("journal.tvalue < 0"))
	dbfilter.append(sqlbuilder.SQLConstant("journal.tday BETWEEN '%s' and '%s'" % (date1, date2)))

	crfilter.append(sqlbuilder.SQLConstant("journal.tvalue > 0"))
	crfilter.append(sqlbuilder.SQLConstant("journal.tday BETWEEN '%s' and '%s'" % (date1, date2)))

	laccounts = {}

	for laccount in orgs.get(org_id).chart.accounts:
		# Rules out undesired types of accounts
		# (aggregate accounts are all accepted, the caller is responsible by filter them out)
		if not laccount.aggregate:
			if result == 1 and laccount.xclass.is_income > 0:
				continue
			elif result == 2 and laccount.xclass.is_income == 0:
				continue

			if normal_balance == 1 and laccount.xclass.normal_balance > 0:
				continue
			elif normal_balance == 2 and laccount.xclass.normal_balance == 0:
				continue

		db = journal.select(sqlbuilder.AND(sqlbuilder.SQLConstant("journal.account_id = %d" % laccount.id), *dbfilter), \
			join=sqlbuilder.INNERJOINOn(None, batches, sqlbuilder.SQLConstant("journal.batch_id = batches.id")))
		cr = journal.select(sqlbuilder.AND(sqlbuilder.SQLConstant("journal.account_id = %d" % laccount.id), *crfilter), \
			join=sqlbuilder.INNERJOINOn(None, batches, sqlbuilder.SQLConstant("journal.batch_id = batches.id")))
		inbal = journal.select(sqlbuilder.AND(sqlbuilder.SQLConstant("journal.account_id = %d" % laccount.id), *inbalfilter), \
			join=sqlbuilder.INNERJOINOn(None, batches, sqlbuilder.SQLConstant("journal.batch_id = batches.id")))

		db_value = db.accumulate('SUM(journal.tvalue)')
		cr_value = cr.accumulate('SUM(journal.tvalue)')
		inbal_value = inbal.accumulate('SUM(journal.tvalue)')

		db_value = db_value and int(db_value) or 0
		cr_value = cr_value and int(cr_value) or 0
		inbal_value = inbal_value and int(inbal_value) or 0

		level = accountlib.account_level(laccount.code)
		parent_class = collapse_class = accountlib.account_to_class(laccount.parent)	
		code_class = accountlib.account_to_class(laccount.code)
		if level <= 1:
			is_leaf = 0
			collapse_class += " collapsed"
		elif not laccount.aggregate:
			is_leaf = 1 
			collapse_class += " invisible"
		else:
			is_leaf = 0
			collapse_class += " collapsed invisible"

		laccounts[laccount.code] = {"code": laccount.code, 
				"account_id": laccount.id, 
				"parent": laccount.parent,
				"description": laccount.description,
				"class": laccount.xclass.description,
				"class_id": laccount.xclass.id,
				"is_income": laccount.xclass.is_income,
				"normal_balance": laccount.xclass.normal_balance,
				"level": accountlib.account_level(laccount.code),
				"aggregate": laccount.aggregate,
				"leaf": is_leaf,
				"code_class": code_class,
				"collapse_class": collapse_class,
				"dirty": 1}

		laccounts[laccount.code]["db"] = db_value
		laccounts[laccount.code]["cr"] = cr_value
		laccounts[laccount.code]["inbal"] = inbal_value
		laccounts[laccount.code]["db_up"] = laccounts[laccount.code]["db"]
		laccounts[laccount.code]["cr_up"] = laccounts[laccount.code]["cr"]
		laccounts[laccount.code]["inbal_up"] = laccounts[laccount.code]["inbal"]

	# Recursively sum uphill to aggregates

	dirty = 1
	while dirty:
		dirty = 0
		for code in laccounts.keys():
			laccount = laccounts[code]
			# if it is dirty and is not top-level, sum uphill
			if (laccount["dirty"] and laccount["parent"]):
				# Sums uphill to parent
				laccounts[laccount["parent"]]["inbal"] += laccount["inbal_up"]
				laccounts[laccount["parent"]]["inbal_up"] += laccount["inbal_up"]
				# Uphill summing is exhausthed
				laccount["inbal_up"] = 0
				
				# Keeping <val> and <val>_up separate allows us to
				# sum uphill on-demand (when some daughter adds to
				# our totals) while retaining the grand total 
				#
				# Otherwise we would have to ensure that grand total
				# of each level is perfectly ready before summing up
				# to the next level. 

				laccounts[laccount["parent"]]["db"] += laccount["db_up"]
				laccounts[laccount["parent"]]["db_up"] += laccount["db_up"]
				laccount["db_up"] = 0

				laccounts[laccount["parent"]]["cr"] += laccount["cr_up"]
				laccounts[laccount["parent"]]["cr_up"] += laccount["cr_up"]
				laccount["cr_up"] = 0
				
				# Parent is dirty (must sum uphill to grandparent)
				laccounts[laccount["parent"]]["dirty"] = 1
				# but we are clean
				laccount["dirty"] = 0
				# Something changed, another round is needed
				dirty = 1

	bal = []

	for code in laccounts.keys():
		account = laccounts[code]
		account["db"] = -account["db"]
		bal.append(account)

	return bal

