#!/usr/bin/env python
# -*- coding: UTF-8 -*-
"""Client for the SOAP interface of foreign exchange rate web service of FRBNY.

Description from http://www.ny.frb.org/markets/pilotfx.html:

The Federal Reserve Bank of New York now offers Noon Foreign Exchange Rates (exchange rates) data as a Web Service. The Noon Foreign Exchange Rates Web Service allows you to query the latest available exchange rates as follows:

    * by date
    * by currency code
    * by currency code with date range

The web service uses SOAP messages to communicate. It exposes the following operations:

    string getAllLatestNoonRates ( )
    returns: latest available Noon exchange rates

    string getAllNoonRates ( date date1 )
    returns: latest available Noon exchange rates for the specified date

    string getLatestNoonRate ( string currency_code )
    returns: latest Noon exchange rate available for the specified currency code

    string getNoonRates ( string currency_code, date date1, date date2 )
    returns: latest Noon exchange rate available for the specified currency code, date range

Returned values are in XML format that conforms to the emerging Statistical Data and Metadata Exchange (SDMX) standard for use in automated processes.
"""

__all__ = [
    'get_all_noon_rates',
    'get_all_latest_noon_rates',
    'get_latest_noon_rate',
    'get_noon_rates',
]

__program__   = "fxrates_client.py"
__author__    = "Christopher Arndt"
__revision__  = "$Rev:$"
__date__      = "$Date:$"
__license__   = "LGPL"

import decimal
import datetime
import sys

try:
    import SOAPpy
    has_soappy = True
except ImportError:
    has_soappy = False

try:
    from xml.etree import cElementTree as ET
except ImportError:
    try:
        import cElementTree as ET
    except ImportError:
        from ElementTree import ElementTree as ET
from xml.parsers.expat import ExpatError

SERVICE_URL = 'http://www.newyorkfed.org/markets/fxrates/WebService/v1_0/FXWS.cfc'
FRBNY_NS = 'http://www.newyorkfed.org/xml/schemas/FX/utility'
ECB_NS = 'http://www.ecb.int/vocabulary/stats/exr'


class ParseError(Exception): pass 
class NetworkError(Exception): pass


_proxy = None
def proxy_factory():
    global _proxy
    if _proxy is None:
        _proxy = SOAPpy.SOAPProxy(SERVICE_URL)
    return _proxy

def convert_date(date):
    if isinstance(date, (SOAPpy.dateType, SOAPpy.dateTimeType)):
        return date
    elif isinstance(date, (datetime.date, datetime.datetime)):
        return SOAPpy.dateType(date.timetuple()[:3])
    elif isinstance(date, (int, float)):
        return SOAPpy.dateType(date)
    raise ValueError('Can not convert date value')

def parse_frbny_sdmx(el):
    """Parse FRBNY variant of SDMX with foreign exchange rates."""

    newctx = decimal.ExtendedContext
    oldctx = decimal.getcontext()
    decimal.setcontext(newctx)

    rates = {}

    for series in el.findall('{%s}Series' % FRBNY_NS):
        key = series.find('{%s}Key' % FRBNY_NS)
        curr = key.find('{%s}CURR' % FRBNY_NS).text.strip()
        switch_curr = True
        if curr == u'USD':
            curr = series.get('UNIT')
            switch_curr = False
        rates[curr] = []
        for obs in series.findall('{%s}Obs' % FRBNY_NS):
            try:
                rate = obs.find('{%s}OBS_VALUE' % FRBNY_NS)
                if rate is None:
                    raise ValueError
                rate = decimal.Decimal(rate.text.strip())
            except ValueError:
                print >>sys.stderr, "Could not parse rate %r: %s" % (rate, e)
                continue
            else:
                if switch_curr:
                    rate = 1 / rate
            try:
                date = obs.find('{%s}TIME_PERIOD' % FRBNY_NS)
                if date is None:
                    raise ValueError
                date = date.text.strip()
                year, month, day = [int(x) for x in date.split('-', 2)]
                date = datetime.date(year, month, day)
            except ValueError, e:
                print >>sys.stderr, "Could not parse date %r: %s" % (date, e)
                continue
            rates[curr].append((rate, date))
    decimal.setcontext(oldctx)
    return rates

def parse_ecb_sdmx(el):
    """Parse ECB variant of SDMX with foreign exchange rates.

    See http://sdw.ecb.int/browse.do?currentNodeId=2018779
    """

    newctx = decimal.ExtendedContext
    oldctx = decimal.getcontext()
    decimal.setcontext(newctx)

    rates = {}

    for series in el.findall('{%s}Series' % ECB_NS):
        curr = series.get('CURRENCY')
        rates[curr] = []
        for obs in series.findall('{%s}Obs' % ECB_NS):
            try:
                rate = obs.get('OBS_VALUE')
                if rate is None:
                    raise ValueError
                elif rate == 'NaN':
                    continue
                rate = decimal.Decimal(rate.strip().replace(',', '.'))
            except ValueError:
                print >>sys.stderr, "Could not parse rate %r: %s" % (rate, e)
                continue
            try:
                date = obs.get('TIME_PERIOD')
                if date is None:
                    raise ValueError
                date = date.strip()
                year, month, day = [int(x) for x in date.split('-', 2)]
                date = datetime.date(year, month, day)
            except ValueError, e:
                print >>sys.stderr, "Could not parse date %r: %s" % (date, e)
                continue
            rates[curr].append((rate, date))
    decimal.setcontext(oldctx)
    return rates


def parse_sdmx(sdmx):
    """Parse SDMX document with currency rates.

    Returns a dictionary with currency codes as keys
    and tuples of (date, rate)-items.
    """

    try:
        doc = ET.fromstring(sdmx)
        el = doc.find('{%s}DataSet' % FRBNY_NS)
        if el:
            return parse_frbny_sdmx(el)
        el = doc.find('{%s}DataSet' % ECB_NS)
        if el:
            return parse_ecb_sdmx(el)
        raise ParseError("Could not find fxrate data in SDMX document.")
    except ExpatError:
        raise ParseError("Could not parse SDMX document.")

def get_all_latest_noon_rates():
    """Returns: latest available Noon exchange rates."""

    if not has_soappy:
        raise NetworkError("No SOAP support present.")
    proxy = proxy_factory()
    try:
        sdmx = proxy.getAllLatestNoonRates()
    except:
        raise NetworkError("SOAP communication error")
    rates = parse_sdmx(sdmx)
    return rates

def get_all_noon_rates(date):
    """Returns: latest available Noon exchange rates for given date."""

    if not has_soappy:
        raise NetworkError("No SOAP support present.")
    proxy = proxy_factory()
    try:
        sdmx = proxy.getAllNoonRates(convert_date(date))
    except:
        raise NetworkError("SOAP communication error")
    rates = parse_sdmx(sdmx)
    return rates

def get_latest_noon_rate(currency):
    """Returns: latest Noon exchange rate available for given currency code."""

    if not has_soappy:
        raise NetworkError("No SOAP support present.")
    proxy = proxy_factory()
    try:
        sdmx = proxy.getLatestNoonRate(currency)
    except:
        raise NetworkError("SOAP communication error")
    rates = parse_sdmx(sdmx)
    return rates[currency][0]

def get_noon_rates(currency, startdate, enddate):
    """Returns: latest Noon exchange rate available for the given currency
    code, and date range.
    """

    if not has_soappy:
        raise NetworkError("No SOAP support present.")
    proxy = proxy_factory()
    try:
        sdmx = proxy.getNoonRates(currency, convert_date(startdate),
          convert_date(enddate))
    except:
        raise NetworkError("SOAP communication error")
    rates = parse_sdmx(sdmx)
    return rates[currency]


def main(args):
    try:
        if args:
            sdmx = open(args[0]).read()
            rates = parse_sdmx(sdmx)
        else:
            rates = get_all_latest_noon_rates()
    except (IOError, OSError), exc:
        print >>sys.stderr, "Could not open %s: %s" % (args[0], exc)
    except KeyboardInterrupt:
        pass
    else:
        from pprint import pprint
        pprint(rates)

    return 0


if __name__ == '__main__':
    sys.exit(main(sys.argv[1:]))
